import torch_scatter


def scatter_(name, src, index, dim_size=None):
    r"""Aggregates all values from the :attr:`src` tensor at the indices
    specified in the :attr:`index` tensor along the first dimension.
    If multiple indices reference the same location, their contributions
    are aggregated according to :attr:`name` (:obj:`"add"`, :obj:`"mean"`,
    :obj:`"max"`).

    Args:
        name (string): The aggregation to use (one of :obj:`"add"`,
            :obj:`"mean"`, :obj:`"max"`).
        src (Tensor): The source tensor.
        index (LongTensor): The indices of elements to scatter.
        dim_size (int, optional): Automatically create output tensor with size
            :attr:`dim_size` in the first dimension. If set to :attr:`None`, a
            minimal sized output tensor is returned. (default: :obj:`None`)

    :rtype: :class:`Tensor`

    .. testsetup::

        import torch

    .. testcode::

        from torch_geometric.utils import scatter_
        src = torch.Tensor([2, 3, -2, 1, 1])
        index = torch.tensor([0, 1, 0, 1, 2])
        output = scatter_("add", src, index)
        print(output)

    .. testoutput::

       tensor([0., 4., 1.])
    """

    assert name in ['add', 'mean', 'max']

    op = getattr(torch_scatter, 'scatter_{}'.format(name))
    fill_value = -1e38 if name is 'max' else 0

    out = op(src, index, 0, None, dim_size, fill_value)
    if isinstance(out, tuple):
        out = out[0]

    if name is 'max':
        out[out == fill_value] = 0

    return out



import torch
from torch_cluster import grid_cluster

from ..repeat import repeat


def voxel_grid(pos, batch, size, start=None, end=None):
    pos = pos.unsqueeze(-1) if pos.dim() == 1 else pos
    num_nodes, dim = pos.size()

    size, start, end = repeat(size, dim), repeat(start, dim), repeat(end, dim)

    pos = torch.cat([pos, batch.unsqueeze(-1).type_as(pos)], dim=-1)
    size = size + [1]
    start = None if start is None else start + [0]
    end = None if end is None else end + [batch.max().item()]

    size = torch.tensor(size, dtype=pos.dtype, device=pos.device)
    if start is not None:
        start = torch.tensor(start, dtype=pos.dtype, device=pos.device)
    if end is not None:
        end = torch.tensor(end, dtype=pos.dtype, device=pos.device)

    return grid_cluster(pos, size, start, end)



import os.path as osp
import shutil

import torch
from torch_geometric.data import InMemoryDataset, extract_zip
from torch_geometric.read import read_ply


class FAUST(InMemoryDataset):
    url = 'http://faust.is.tue.mpg.de/'

    def __init__(self,
                 root,
                 train=True,
                 transform=None,
                 pre_transform=None,
                 pre_filter=None):
        super(FAUST, self).__init__(root, transform, pre_transform, pre_filter)
        path = self.processed_paths[0] if train else self.processed_paths[1]
        self.data, self.slices = torch.load(path)

    @property
    def raw_file_names(self):
        return 'MPI-FAUST.zip'

    @property
    def processed_file_names(self):
        return ['training.pt', 'test.pt']

    def download(self):
        raise RuntimeError(
            'Dataset not found. Please download MPI-FAUST.zip from {} and '
            'move it to {}'.format(self.url, self.raw_dir))

    def process(self):
        extract_zip(self.raw_paths[0], self.raw_dir, log=False)

        path = osp.join(self.raw_dir, 'MPI-FAUST', 'training', 'registrations')
        path = osp.join(path, 'tr_reg_{0:03d}.ply')
        data_list = []
        for i in range(100):
            data = read_ply(path.format(i))
            data.y = torch.tensor([i % 10], dtype=torch.long)
            if self.pre_filter is not None and not self.pre_filter(data):
                continue
            if self.pre_transform is not None:
                data = self.pre_transform(data)
            data_list.append(data)

        torch.save(self.collate(data_list[:80]), self.processed_paths[0])
        torch.save(self.collate(data_list[80:]), self.processed_paths[1])

        shutil.rmtree(osp.join(self.raw_dir, 'MPI-FAUST'))



# def test_collate_to_set():
#     x1 = torch.tensor([1, 2, 3], dtype=torch.float)
#     e1 = torch.tensor([[0, 1, 1, 2], [1, 0, 2, 1]])
#     x2 = torch.tensor([1, 2], dtype=torch.float)
#     e2 = torch.tensor([[0, 1], [1, 0]])

#     data, slices = collate_to_set([Data(x1, e1), Data(x2, e2)])

#     assert len(data) == 2
#     assert data.x.tolist() == [1, 2, 3, 1, 2]
#     data.edge_index.tolist() == [[0, 1, 1, 2, 0, 1], [1, 0, 2, 1, 1, 0]]
#     assert len(slices.keys()) == 2
#     assert slices['x'].tolist() == [0, 3, 5]
#     assert slices['edge_index'].tolist() == [0, 4, 6]



from torch_scatter import scatter_max
from torch_geometric.data import Batch

from .consecutive import consecutive_cluster
from .pool import pool_edge, pool_batch, pool_pos


def _max_pool_x(cluster, x, size=None):
    fill = -9999999
    x, _ = scatter_max(x, cluster, dim=0, dim_size=size, fill_value=fill)
    x[x == fill] = 0
    return x


def max_pool_x(cluster, x, batch, size=None):
    if size is not None:
        return _max_pool_x(cluster, x, (batch.max().item() + 1) * size)

    cluster, perm = consecutive_cluster(cluster)
    x = _max_pool_x(cluster, x)
    batch = pool_batch(perm, batch)

    return x, batch


def max_pool(cluster, data, transform=None):
    cluster, perm = consecutive_cluster(cluster)

    x = _max_pool_x(cluster, data.x)
    index, attr = pool_edge(cluster, data.edge_index, data.edge_attr)
    batch = None if data.batch is None else pool_batch(perm, data.batch)
    pos = None if data.pos is None else pool_pos(cluster, data.pos)

    data = Batch(batch=batch, x=x, edge_index=index, edge_attr=attr, pos=pos)

    if transform is not None:
        data = transform(data)

    return data



from torch_sparse import coalesce


def dense_to_sparse(tensor):
    index = tensor.nonzero()
    value = tensor[index]
    index = index.t().contiguous()
    index, value = coalesce(index, value, tensor.size(0), tensor.size(1))
    return index, value



import torch


def accuracy(pred, target):
    return (pred == target).sum().item() / target.numel()


def true_positive(pred, target, num_classes):
    out = []
    for i in range(num_classes):
        out.append(((pred == i) & (target == i)).sum())

    return torch.tensor(out)


def true_negative(pred, target, num_classes):
    out = []
    for i in range(num_classes):
        out.append(((pred != i) & (target != i)).sum())

    return torch.tensor(out)


def false_positive(pred, target, num_classes):
    out = []
    for i in range(num_classes):
        out.append(((pred == i) & (target != i)).sum())

    return torch.tensor(out)


def false_negative(pred, target, num_classes):
    out = []
    for i in range(num_classes):
        out.append(((pred != i) & (target == i)).sum())

    return torch.tensor(out)


def precision(pred, target, num_classes):
    tp = true_positive(pred, target, num_classes).to(torch.float)
    fp = false_positive(pred, target, num_classes).to(torch.float)

    out = tp / (tp + fp)
    out[torch.isnan(out)] = 0

    return out


def recall(pred, target, num_classes):
    tp = true_positive(pred, target, num_classes).to(torch.float)
    fn = false_negative(pred, target, num_classes).to(torch.float)

    out = tp / (tp + fn)
    out[torch.isnan(out)] = 0

    return out


def f1_score(pred, target, num_classes):
    prec = precision(pred, target, num_classes)
    rec = recall(pred, target, num_classes)

    score = 2 * (prec * rec) / (prec + rec)
    score[torch.isnan(score)] = 0

    return score



import torch
from torch_geometric.transforms import Cartesian
from torch_geometric.data import Data


def test_cartesian():
    assert Cartesian().__repr__() == 'Cartesian(norm=True, max_value=None)'

    pos = torch.tensor([[-1, 0], [0, 0], [2, 0]], dtype=torch.float)
    edge_index = torch.tensor([[0, 1, 1, 2], [1, 0, 2, 1]])
    data = Data(edge_index=edge_index, pos=pos)

    out = Cartesian()(data).edge_attr.tolist()
    assert out == [[0.75, 0.5], [0.25, 0.5], [1, 0.5], [0, 0.5]]

    out = Cartesian(max_value=4, cat=False)(data).edge_attr.tolist()
    assert out == [[0.625, 0.5], [0.375, 0.5], [0.75, 0.5], [0.25, 0.5]]

    out = Cartesian(norm=False, cat=False)(data).edge_attr.tolist()
    assert out == [[1, 0], [-1, 0], [2, 0], [-2, 0]]

    data.edge_attr = torch.tensor([1, 1, 1, 1], dtype=torch.float)

    out = Cartesian(cat=True)(data).edge_attr.tolist()
    assert out == [[1, 0.75, 0.5], [1, 0.25, 0.5], [1, 1, 0.5], [1, 0, 0.5]]



from torch_geometric.utils import grid


def test_grid():
    (row, col), pos = grid(height=3, width=2)

    expected_row = [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2]
    expected_col = [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 4, 5]
    expected_row += [3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5]
    expected_col += [0, 1, 2, 3, 4, 5, 2, 3, 4, 5, 2, 3, 4, 5]

    expected_pos = [[0, 2], [1, 2], [0, 1], [1, 1], [0, 0], [1, 0]]

    assert row.tolist() == expected_row
    assert col.tolist() == expected_col
    assert pos.tolist() == expected_pos



import torch
import torch_geometric.transforms as T
from torch_geometric.data import Data


def test_compose():
    transform = T.Compose([T.Center(), T.AddSelfLoops()])
    assert transform.__repr__() == ('Compose([\n'
                                    '    Center(),\n'
                                    '    AddSelfLoops(),\n'
                                    '])')

    pos = torch.tensor([[0, 0], [2, 0], [4, 0]], dtype=torch.float)
    edge_index = torch.tensor([[0, 1, 1, 2], [1, 0, 2, 1]])
    data = Data(edge_index=edge_index, pos=pos)

    out = transform(data)
    assert out.pos.tolist() == [[-2, 0], [0, 0], [2, 0]]
    assert out.edge_index.tolist() == [[0, 0, 1, 1, 1, 2, 2],
                                       [0, 1, 0, 1, 2, 1, 2]]



from __future__ import division

import sys
import random
import os.path as osp
import shutil

import pytest
import torch
from torch_geometric.datasets import TUDataset
from torch_geometric.data import DataLoader, DenseDataLoader
from torch_geometric.transforms import ToDense


def test_enzymes():
    root = osp.join('/', 'tmp', str(random.randrange(sys.maxsize)), 'test')
    dataset = TUDataset(root, 'ENZYMES')

    assert len(dataset) == 600
    assert dataset.num_features == 21
    assert dataset.num_classes == 6
    assert dataset.__repr__() == 'ENZYMES(600)'

    assert len(dataset[0]) == 3
    assert len(dataset.shuffle()) == 600
    assert len(dataset[:100]) == 100
    assert len(dataset[torch.arange(100, dtype=torch.long)]) == 100
    mask = torch.zeros(600, dtype=torch.uint8)
    mask[:100] = 1
    assert len(dataset[mask]) == 100

    loader = DataLoader(dataset, batch_size=len(dataset))
    for data in loader:
        assert data.num_graphs == 600

        avg_num_nodes = data.num_nodes / data.num_graphs
        assert pytest.approx(avg_num_nodes, abs=1e-2) == 32.63

        avg_num_edges = data.num_edges / (2 * data.num_graphs)
        assert pytest.approx(avg_num_edges, abs=1e-2) == 62.14

        assert len(data) == 4
        assert list(data.x.size()) == [data.num_nodes, 21]
        assert list(data.y.size()) == [data.num_graphs]
        assert data.y.max() + 1 == 6
        assert list(data.batch.size()) == [data.num_nodes]

        assert data.contains_isolated_nodes()
        assert not data.contains_self_loops()
        assert data.is_undirected()

    dataset.transform = ToDense(num_nodes=126)
    loader = DenseDataLoader(dataset, batch_size=len(dataset))
    for data in loader:
        assert len(data) == 4
        assert list(data.x.size()) == [600, 126, 21]
        assert list(data.adj.size()) == [600, 126, 126]
        assert list(data.mask.size()) == [600, 126]
        assert list(data.y.size()) == [600, 1]

    shutil.rmtree(root)



import torch
from torch_geometric.transforms import TargetIndegree
from torch_geometric.data import Data


def test_target_indegree():
    assert TargetIndegree().__repr__() == 'TargetIndegree(cat=True)'

    edge_index = torch.tensor([[0, 1, 1, 2], [1, 0, 2, 1]])
    data = Data(edge_index=edge_index)

    out = TargetIndegree()(data).edge_attr.tolist()
    assert out == [[1], [0.5], [0.5], [1]]

    data.edge_attr = torch.tensor([1, 1, 1, 1], dtype=torch.float)
    out = TargetIndegree()(data).edge_attr.tolist()
    assert out == [[1, 1], [1, 0.5], [1, 0.5], [1, 1]]



import torch


def one_hot(src, num_classes=None, dtype=None):
    src = src.to(torch.long)
    src = src.unsqueeze(-1) if src.dim() == 1 else src
    assert src.dim() == 2

    if num_classes is None:
        num_classes = src.max(dim=0)[0] + 1
    elif isinstance(num_classes, int) or isinstance(num_classes, float):
        num_classes = torch.tensor(int(num_classes))

    if src.size(1) > 1:
        zero = torch.tensor([0], device=src.device)
        src = src + torch.cat([zero, torch.cumsum(num_classes, 0)[:-1]])

    size = src.size(0), num_classes.sum()
    out = torch.zeros(size, dtype=dtype, device=src.device)
    out.scatter_(1, src, 1)
    return out



import os
import os.path as osp
import json

import torch
import numpy as np
from torch_sparse import coalesce
from torch_geometric.data import (InMemoryDataset, Data, download_url,
                                  extract_zip)


class PPI(InMemoryDataset):
    url = 'http://snap.stanford.edu/graphsage/ppi.zip'

    def __init__(self, root, transform=None, pre_transform=None):
        super(PPI, self).__init__(root, transform, pre_transform)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def raw_file_names(self):
        prefix = self.__class__.__name__.lower()
        suffix = ['G.json', 'feats.npy', 'class_map.json']
        return ['{}-{}'.format(prefix, s) for s in suffix]

    @property
    def processed_file_names(self):
        return 'data.pt'

    def download(self):
        path = download_url(self.url, self.root)
        extract_zip(path, self.root)
        os.unlink(path)
        name = self.__class__.__name__.lower()
        os.rename(osp.join(self.root, name), self.raw_dir)

    def process(self):
        with open(self.raw_paths[0], 'r') as f:
            graph_data = json.load(f)

        mask = torch.zeros(len(graph_data['nodes']), dtype=torch.uint8)
        for i in graph_data['nodes']:
            mask[i['id']] = 1 if i['val'] else (2 if i['test'] else 0)
        train_mask, val_mask, test_mask = mask == 0, mask == 1, mask == 2

        row, col = [], []
        for i in graph_data['links']:
            row.append(i['source'])
            col.append(i['target'])
        edge_index = torch.stack([torch.tensor(row), torch.tensor(col)], dim=0)
        edge_index, _ = coalesce(edge_index, None, mask.size(0), mask.size(0))

        x = torch.from_numpy(np.load(self.raw_paths[1])).float()

        with open(self.raw_paths[2], 'r') as f:
            y_data = json.load(f)

        y = []
        for i in range(len(y_data)):
            y.append(y_data[str(i)])
        y = torch.tensor(y, dtype=torch.float)

        data = Data(x=x, edge_index=edge_index, y=y)
        data.train_mask = train_mask
        data.val_mask = val_mask
        data.test_mask = test_mask

        data = data if self.pre_transform is None else self.pre_transform(data)
        data, slices = self.collate([data])
        torch.save((data, slices), self.processed_paths[0])

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)



import os.path as osp
from glob import glob

import torch
from torch_geometric.data import InMemoryDataset, extract_zip
from torch_geometric.read import read_ply


class CoMA(InMemoryDataset):
    url = 'https://coma.is.tue.mpg.de/'

    categories = [
        'bareteeth',
        'cheeks_in',
        'eyebrow',
        'high_smile',
        'lips_back',
        'lips_up',
        'mouth_down',
        'mouth_extreme',
        'mouth_middle',
        'mouth_open',
        'mouth_side',
        'mouth_up',
    ]

    def __init__(self,
                 root,
                 train=True,
                 transform=None,
                 pre_transform=None,
                 pre_filter=None):
        super(CoMA, self).__init__(root, transform, pre_transform, pre_filter)
        path = self.processed_paths[0] if train else self.processed_paths[1]
        self.data, self.slices = torch.load(path)

    @property
    def raw_file_names(self):
        return 'COMA_data.zip'

    @property
    def processed_file_names(self):
        return ['training.pt', 'test.pt']

    def download(self):
        raise RuntimeError(
            'Dataset not found. Please download COMA_data.zip from {} and '
            'move it to {}'.format(self.url, self.raw_dir))

    def process(self):
        folders = sorted(glob(osp.join(self.raw_dir, 'FaceTalk_*')))
        if len(folders) == 0:
            extract_zip(self.raw_paths[0], self.raw_dir, log=False)
            folders = sorted(glob(osp.join(self.raw_dir, 'FaceTalk_*')))

        train_data_list, test_data_list = [], []
        for folder in folders:
            for i, category in enumerate(self.categories):
                files = sorted(glob(osp.join(folder, category, '*.ply')))
                for j, f in enumerate(files):
                    data = read_ply(f)
                    data.y = torch.tensor([i], dtype=torch.long)
                    if self.pre_filter is not None and\
                       not self.pre_filter(data):
                        continue
                    if self.pre_transform is not None:
                        data = self.pre_transform(data)

                    if (j % 100) < 90:
                        train_data_list.append(data)
                    else:
                        test_data_list.append(data)

        torch.save(self.collate(train_data_list), self.processed_paths[0])
        torch.save(self.collate(test_data_list), self.processed_paths[1])



import torch

from .num_nodes import maybe_num_nodes
from .loop import remove_self_loops


def contains_isolated_nodes(edge_index, num_nodes=None):
    num_nodes = maybe_num_nodes(edge_index, num_nodes)
    (row, _), _ = remove_self_loops(edge_index)
    return torch.unique(row).size(0) < num_nodes



import torch
from torch_sparse import coalesce

from .num_nodes import maybe_num_nodes


def is_undirected(edge_index, num_nodes=None):
    num_nodes = maybe_num_nodes(edge_index, num_nodes)
    edge_index, _ = coalesce(edge_index, None, num_nodes, num_nodes)
    undirected_edge_index = to_undirected(edge_index, num_nodes=num_nodes)
    return edge_index.size(1) == undirected_edge_index.size(1)


def to_undirected(edge_index, num_nodes=None):
    num_nodes = maybe_num_nodes(edge_index, num_nodes)

    row, col = edge_index
    row, col = torch.cat([row, col], dim=0), torch.cat([col, row], dim=0)
    edge_index = torch.stack([row, col], dim=0)
    edge_index, _ = coalesce(edge_index, None, num_nodes, num_nodes)

    return edge_index



# x1 = torch.tensor([1, 2, 3], dtype=torch.float)
# edge_index1 = torch.tensor([[0, 1, 1, 2], [1, 0, 2, 1]])
# x2 = torch.tensor([1, 2], dtype=torch.float)
# edge_index2 = torch.tensor([[0, 1], [1, 0]])
# data1, data2 = Data(x1, edge_index1), Data(x2, edge_index2)
# dataset, slices = collate_to_set([data1, data2])

# def test_data_from_set():
#     data = data_from_set(dataset, slices, 0)
#     assert len(data) == 2
#     assert data.x.tolist() == x1.tolist()
#     assert data.edge_index.tolist() == edge_index1.tolist()

#     data = data_from_set(dataset, slices, 1)
#     assert len(data) == 2
#     assert data.x.tolist() == x2.tolist()
#     assert data.edge_index.tolist() == edge_index2.tolist()

# def test_split_set():
#     output, output_slices = split_set(dataset, slices, torch.tensor([0]))

#     assert len(output) == 2
#     assert output.x.tolist() == x1.tolist()
#     assert output.edge_index.tolist() == edge_index1.tolist()

#     assert len(output_slices.keys()) == 2
#     assert output_slices['x'].tolist() == [0, 3]
#     assert output_slices['edge_index'].tolist() == [0, 4]



import os
import os.path as osp

import torch
from torch_geometric.data import InMemoryDataset, download_url, extract_zip
from torch_geometric.read import read_tu_data


class TUDataset(InMemoryDataset):
    url = 'https://ls11-www.cs.uni-dortmund.de/people/morris/' \
          'graphkerneldatasets'

    def __init__(self,
                 root,
                 name,
                 transform=None,
                 pre_transform=None,
                 pre_filter=None):
        self.name = name
        super(TUDataset, self).__init__(root, transform, pre_transform,
                                        pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def raw_file_names(self):
        names = ['A', 'graph_indicator']
        return ['{}_{}.txt'.format(self.name, name) for name in names]

    @property
    def processed_file_names(self):
        return 'data.pt'

    def download(self):
        path = download_url('{}/{}.zip'.format(self.url, self.name), self.root)
        extract_zip(path, self.root)
        os.unlink(path)
        os.rename(osp.join(self.root, self.name), self.raw_dir)

    def process(self):
        self.data, self.slices = read_tu_data(self.raw_dir, self.name)

        if self.pre_filter is not None:
            data_list = [self.get(idx) for idx in range(len(self))]
            data_list = [data for data in data_list if self.pre_filter(data)]
            self.data, self.slices = self.collate(data_list)

        if self.pre_transform is not None:
            data_list = [self.get(idx) for idx in range(len(self))]
            data_list = [self.pre_transform(data) for data in data_list]
            self.data, self.slices = self.collate(data_list)

        torch.save((self.data, self.slices), self.processed_paths[0])

    def __repr__(self):
        return '{}({})'.format(self.name, len(self))



import torch
from torch.nn import Parameter
from torch_sparse import spmm
from torch_scatter import scatter_add
from torch_geometric.utils import remove_self_loops, add_self_loops

from ..inits import glorot, zeros


class GCNConv(torch.nn.Module):
    r"""The graph convolutional operator from the `"Semi-supervised
    Classfication with Graph Convolutional Networks"
    <https://arxiv.org/abs/1609.02907>`_ paper

    .. math::
        \mathbf{X}^{\prime} = \mathbf{\hat{D}}^{-1/2} \mathbf{\hat{A}}
        \mathbf{\hat{D}}^{-1/2} \mathbf{X} \mathbf{\Theta},

    where :math:`\mathbf{\hat{A}} = \mathbf{A} + \mathbf{I}` denotes the
    adjacency matrix with inserted self-loops and
    :math:`\hat{D}_{ii} = \sum_{j=0} \hat{A}_{ij}` its diagonal degree matrix.

    Args:
        in_channels (int): Size of each input sample.
        out_channels (int): Size of each output sample.
        improved (bool, optional): If set to :obj:`True`, the layer computes
            :math:`\mathbf{\hat{A}}` as :math:`\mathbf{A} + 2\mathbf{I}`.
            (default: :obj:`False`)
        bias (bool, optional): If set to :obj:`False`, the layer will not learn
            an additive bias. (default: :obj:`True`)
    """

    def __init__(self, in_channels, out_channels, improved=False, bias=True):
        super(GCNConv, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.improved = improved

        self.weight = Parameter(torch.Tensor(in_channels, out_channels))

        if bias:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        glorot(self.weight)
        zeros(self.bias)

    def forward(self, x, edge_index, edge_attr=None):
        """"""
        x = x.unsqueeze(-1) if x.dim() == 1 else x

        edge_index, edge_attr = remove_self_loops(edge_index, edge_attr)
        if edge_attr is None:
            edge_attr = x.new_ones((edge_index.size(1), ))
        assert edge_attr.dim() == 1 and edge_attr.numel() == edge_index.size(1)

        # Add self-loops to adjacency matrix.
        edge_index = add_self_loops(edge_index, x.size(0))
        loop_value = x.new_full((x.size(0), ), 1 if not self.improved else 2)
        edge_attr = torch.cat([edge_attr, loop_value], dim=0)

        # Normalize adjacency matrix.
        row, col = edge_index
        deg = scatter_add(edge_attr, row, dim=0, dim_size=x.size(0))
        deg = deg.pow(-0.5)
        deg[deg == float('inf')] = 0
        edge_attr = deg[row] * edge_attr * deg[col]

        # Perform the convolution.
        out = torch.mm(x, self.weight)
        out = spmm(edge_index, edge_attr, out.size(0), out)

        if self.bias is not None:
            out = out + self.bias

        return out

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__, self.in_channels,
                                   self.out_channels)



import torch
from torch_geometric.utils import is_undirected, to_undirected


def test_is_undirected():
    row = torch.tensor([0, 1, 0])
    col = torch.tensor([1, 0, 0])

    assert is_undirected(torch.stack([row, col], dim=0))

    row = torch.tensor([0, 1, 1])
    col = torch.tensor([1, 0, 2])

    assert not is_undirected(torch.stack([row, col], dim=0))


def test_to_undirected():
    row = torch.tensor([0, 1, 1])
    col = torch.tensor([1, 0, 2])

    edge_index = to_undirected(torch.stack([row, col], dim=0))
    assert edge_index.tolist() == [[0, 1, 1, 2], [1, 0, 2, 1]]



import torch
import scipy.sparse
import networkx as nx

from .num_nodes import maybe_num_nodes


def to_scipy_sparse_matrix(edge_index, edge_attr=None, num_nodes=None):
    row, col = edge_index.cpu()

    if edge_attr is None:
        edge_attr = torch.ones(row.size(0))
    else:
        edge_attr = edge_attr.view(-1).cpu()
        assert edge_attr.size(0) == row.size(0)

    N = maybe_num_nodes(edge_index, num_nodes)
    out = scipy.sparse.coo_matrix((edge_attr, (row, col)), (N, N))
    return out


def to_networkx(edge_index, x=None, edge_attr=None, pos=None, num_nodes=None):
    num_nodes = num_nodes if x is None else x.size(0)
    num_nodes = num_nodes if pos is None else pos.size(0)
    num_nodes = maybe_num_nodes(edge_index, num_nodes)

    G = nx.Graph()

    for i in range(num_nodes):
        G.add_node(i)
        if x is not None:
            G.nodes[i]['x'] = x[i].cpu().numpy()
        if pos is not None:
            G.nodes[i]['pos'] = pos[i].cpu().numpy()

    for i in range(edge_index.size(1)):
        source, target = edge_index[0][i].item(), edge_index[1][i].item()
        G.add_edge(source, target)
        if edge_attr is not None:
            if edge_attr.numel() == edge_attr.size(0):
                G[source][target]['weight'] = edge_attr[i].item()
            else:
                G[source][target]['weight'] = edge_attr[i].cpu().numpy()

    return G



import os

import torch
from torch_geometric.data import (InMemoryDataset, download_url, extract_tar,
                                  Data)


class QM9(InMemoryDataset):
    url = 'http://www.roemisch-drei.de/qm9.tar.gz'

    def __init__(self,
                 root,
                 transform=None,
                 pre_transform=None,
                 pre_filter=None):
        super(QM9, self).__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def raw_file_names(self):
        return 'qm9.pt'

    @property
    def processed_file_names(self):
        return 'data.pt'

    def download(self):
        file_path = download_url(self.url, self.raw_dir)
        extract_tar(file_path, self.raw_dir, mode='r')
        os.unlink(file_path)

    def process(self):
        raw_data_list = torch.load(self.raw_paths[0])
        data_list = [
            Data(
                x=d['x'],
                edge_index=d['edge_index'],
                edge_attr=d['edge_attr'],
                y=d['y'],
                pos=d['pos']) for d in raw_data_list
        ]

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[0])



import torch
from torch_scatter import scatter_add


def to_batch(x, batch, fill_value=0):
    num_nodes = scatter_add(batch.new_ones(x.size(0)), batch, dim=0)
    batch_size, max_num_nodes = num_nodes.size(0), num_nodes.max().item()
    cum_nodes = torch.cat([batch.new_zeros(1), num_nodes.cumsum(dim=0)], dim=0)

    index = torch.arange(batch.size(0), dtype=torch.long, device=x.device)
    index = (index - cum_nodes[batch]) + (batch * max_num_nodes)

    size = [batch_size * max_num_nodes] + list(x.size())[1:]
    batch_x = x.new_full(size, fill_value)
    batch_x[index] = x
    size = [batch_size, max_num_nodes] + list(x.size())[1:]
    batch_x = batch_x.view(size)

    return batch_x, num_nodes



import torch
from torch_geometric.transforms import RandomFlip
from torch_geometric.data import Data


def test_cartesian():
    assert RandomFlip(axis=0).__repr__() == 'RandomFlip(axis=0, p=0.5)'

    pos = torch.tensor([[-1, 1], [-3, 0], [2, -1]], dtype=torch.float)
    data = Data(pos=pos)

    out = RandomFlip(axis=0, p=1)(data).pos.tolist()
    assert out == [[1, 1], [3, 0], [-2, -1]]



import torch
from torch_geometric.transforms import LinearTransformation
from torch_geometric.data import Data


def test_cartesian():
    matrix = torch.tensor([[2, 0], [0, 2]], dtype=torch.float)
    transform = LinearTransformation(matrix)
    out = 'LinearTransformation([[2.0, 0.0], [0.0, 2.0]])'
    assert transform.__repr__() == out

    pos = torch.tensor([[-1, 1], [-3, 0], [2, -1]], dtype=torch.float)
    data = Data(pos=pos)

    out = transform(data).pos.tolist()
    assert out == [[-2, 2], [-6, 0], [4, -2]]



import torch
from torch_geometric.utils import (contains_isolated_nodes,
                                   contains_self_loops, is_undirected)

from ..utils.num_nodes import maybe_num_nodes


class Data(object):
    """"""

    def __init__(self,
                 x=None,
                 edge_index=None,
                 edge_attr=None,
                 y=None,
                 pos=None):
        self.x = x
        self.edge_index = edge_index
        self.edge_attr = edge_attr
        self.y = y
        self.pos = pos

    @staticmethod
    def from_dict(dictionary):
        data = Data()
        for key, item in dictionary.items():
            data[key] = item
        return data

    def __getitem__(self, key):
        """"""
        return getattr(self, key)

    def __setitem__(self, key, item):
        """"""
        setattr(self, key, item)

    @property
    def keys(self):
        """"""
        return [key for key in self.__dict__.keys() if self[key] is not None]

    def __len__(self):
        """"""
        return len(self.keys)

    def __contains__(self, key):
        """"""
        return key in self.keys

    def __iter__(self):
        """"""
        for key in sorted(self.keys):
            yield key, self[key]

    def __call__(self, *keys):
        for key in sorted(self.keys) if not keys else keys:
            if self[key] is not None:
                yield key, self[key]

    def cat_dim(self, key, item):
        return -1 if item.dtype == torch.long else 0

    @property
    def num_nodes(self):
        """"""
        for key, item in self('x', 'pos'):
            return item.size(self.cat_dim(key, item))
        if self.edge_index is not None:
            return maybe_num_nodes(self.edge_index)
        return None

    @property
    def num_edges(self):
        """"""
        for key, item in self('edge_index', 'edge_attr'):
            return item.size(self.cat_dim(key, item))
        return None

    @property
    def num_features(self):
        """"""
        return 1 if self.x.dim() == 1 else self.x.size(1)

    def is_coalesced(self):
        """"""
        row, col = self.edge_index
        index = self.num_nodes * row + col
        return row.size(0) == torch.unique(index).size(0)

    def contains_isolated_nodes(self):
        """"""
        return contains_isolated_nodes(self.edge_index, self.num_nodes)

    def contains_self_loops(self):
        """"""
        return contains_self_loops(self.edge_index)

    def is_undirected(self):
        """"""
        return is_undirected(self.edge_index, self.num_nodes)

    def is_directed(self):
        """"""
        return not self.is_undirected()

    def apply(self, func, *keys):
        for key, item in self(*keys):
            self[key] = func(item)
        return self

    def contiguous(self, *keys):
        """"""
        return self.apply(lambda x: x.contiguous(), *keys)

    def to(self, device, *keys):
        """"""
        return self.apply(lambda x: x.to(device), *keys)

    def __repr__(self):
        info = ['{}={}'.format(key, list(item.size())) for key, item in self]
        return '{}({})'.format(self.__class__.__name__, ', '.join(info))



import torch
from torch_geometric.utils import degree, one_hot


class OneHotDegree(object):
    r"""Adds the node degree as one hot encodings to the node features.

    Args:
        max_degree (int): Maximum degree.
        cat (bool, optional): Concat node degrees to node features instead
            of replacing them. (default: :obj:`True`)

    .. testsetup::

        import torch
        from torch_geometric.data import Data

    .. testcode::

        from torch_geometric.transforms import OneHotDegree

        edge_index = torch.tensor([[0, 1, 1, 2], [1, 0, 2, 1]])
        data = Data(edge_index=edge_index)

        data = OneHotDegree(max_degree=2)(data)

        print(data.x)

    .. testoutput::

        tensor([[0., 1., 0.],
                [0., 0., 1.],
                [0., 1., 0.]])
    """

    def __init__(self, max_degree, cat=True):
        self.max_degree = max_degree
        self.cat = cat

    def __call__(self, data):
        row, x = data.edge_index[0], data.x
        deg = degree(row, data.num_nodes)
        deg = one_hot(deg, num_classes=self.max_degree + 1)

        if x is not None and self.cat:
            x = x.view(-1, 1) if x.dim() == 1 else x
            data.x = torch.cat([x, deg.to(x.dtype)], dim=-1)
        else:
            data.x = deg

        return data

    def __repr__(self):
        return '{}({}, cat={})'.format(self.__class__.__name__,
                                       self.max_degree, self.cat)



from torch_geometric.utils import degree


def normalized_cut(edge_index, edge_attr, num_nodes=None):
    row, col = edge_index
    deg = 1 / degree(row, num_nodes, edge_attr.dtype)
    deg = deg[row] + deg[col]
    cut = edge_attr * deg
    return cut



import torch
from torch_geometric.utils import contains_isolated_nodes


def test_contains_isolated_nodes():
    row = torch.tensor([0, 1, 0])
    col = torch.tensor([1, 0, 0])

    assert not contains_isolated_nodes(torch.stack([row, col], dim=0))
    assert contains_isolated_nodes(torch.stack([row, col], dim=0), 3)

    row = torch.tensor([0, 1, 2, 0])
    col = torch.tensor([1, 0, 2, 0])
    assert contains_isolated_nodes(torch.stack([row, col], dim=0))



import collections
import os.path as osp

import torch.utils.data

from .makedirs import makedirs


def to_list(x):
    if not isinstance(x, collections.Iterable) or isinstance(x, str):
        x = [x]
    return x


def files_exist(files):
    return all([osp.exists(f) for f in files])


class Dataset(torch.utils.data.Dataset):
    @property
    def raw_file_names(self):
        """"""
        raise NotImplementedError

    @property
    def processed_file_names(self):
        """"""
        raise NotImplementedError

    def download(self):
        """"""
        raise NotImplementedError

    def process(self):
        """"""
        raise NotImplementedError

    def __len__(self):
        """"""
        raise NotImplementedError

    def get(self, idx):
        """"""
        raise NotImplementedError

    def __init__(self,
                 root,
                 transform=None,
                 pre_transform=None,
                 pre_filter=None):
        super(Dataset, self).__init__()

        self.root = osp.expanduser(osp.normpath(root))
        self.raw_dir = osp.join(self.root, 'raw')
        self.processed_dir = osp.join(self.root, 'processed')
        self.transform = transform
        self.pre_transform = pre_transform
        self.pre_filter = pre_filter

        self._download()
        self._process()

    @property
    def num_features(self):
        """"""
        return self[0].num_features

    @property
    def raw_paths(self):
        files = to_list(self.raw_file_names)
        return [osp.join(self.raw_dir, f) for f in files]

    @property
    def processed_paths(self):
        files = to_list(self.processed_file_names)
        return [osp.join(self.processed_dir, f) for f in files]

    def _download(self):
        if files_exist(self.raw_paths):  # pragma: no cover
            return

        makedirs(self.raw_dir)
        self.download()

    def _process(self):
        if files_exist(self.processed_paths):  # pragma: no cover
            return

        print('Processing...')

        makedirs(self.processed_dir)
        self.process()

        print('Done!')

    def __getitem__(self, idx):
        data = self.get(idx)
        data = data if self.transform is None else self.transform(data)
        return data

    def __repr__(self):  # pragma: no cover
        return '{}({})'.format(self.__class__.__name__, len(self))



import torch
from torch_geometric.transforms import SamplePoints
from torch_geometric.data import Data


def test_sample_points():
    assert SamplePoints(1024).__repr__() == 'SamplePoints(1024)'

    pos = [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0]]
    data = Data(pos=torch.tensor(pos, dtype=torch.float))
    data.face = torch.tensor([[0, 1], [1, 2], [2, 3]], dtype=torch.long)

    data = SamplePoints(8)(data)
    pos = data.pos
    assert pos[:, 0].min().item() >= 0 and pos[:, 0].max().item() <= 1
    assert pos[:, 1].min().item() >= 0 and pos[:, 1].max().item() <= 1
    assert pos[:, 2].abs().sum().item() == 0
    assert 'face' not in data



import torch

from .num_nodes import maybe_num_nodes


def contains_self_loops(edge_index):
    row, col = edge_index
    mask = row == col
    return mask.sum().item() > 0


def remove_self_loops(edge_index, edge_attr=None):
    row, col = edge_index
    mask = row != col
    edge_attr = edge_attr if edge_attr is None else edge_attr[mask]
    mask = mask.unsqueeze(0).expand_as(edge_index)
    edge_index = edge_index[mask].view(2, -1)

    return edge_index, edge_attr


def add_self_loops(edge_index, num_nodes=None):
    num_nodes = maybe_num_nodes(edge_index, num_nodes)

    dtype, device = edge_index.dtype, edge_index.device
    loop = torch.arange(0, num_nodes, dtype=dtype, device=device)
    loop = loop.unsqueeze(0).repeat(2, 1)
    edge_index = torch.cat([edge_index, loop], dim=1)

    return edge_index



from torch_scatter import scatter_max, scatter_add

from .num_nodes import maybe_num_nodes


def softmax(src, index, num_nodes=None):
    num_nodes = maybe_num_nodes(index, num_nodes)

    out = src - scatter_max(src, index, dim=0, dim_size=num_nodes)[0][index]
    out = out.exp()
    out = out / scatter_add(out, index, dim=0, dim_size=num_nodes)[index]

    return out



# import torch
# from torch_geometric.nn.pool.voxel_grid import voxel_grid

# def test_voxel_grid():
#     pos = torch.Tensor([[0.5, 0.5], [1.2, 1.5], [0.2, 1.7], [0.3, 0.4],
#                         [0.5, 0.5], [1.2, 1.5], [0.2, 1.7], [0.3, 0.4]])
#     batch = torch.LongTensor([0, 0, 0, 0, 1, 1, 1, 1])

#     cluster, batch = voxel_grid(pos, size=1, start=0, end=2, batch=batch)
#     assert cluster.tolist() == [0, 2, 1, 0, 3, 5, 4, 3]
#     assert batch.tolist() == [0, 0, 0, 1, 1, 1]



import torch
from torch_geometric.data import Data, Batch


def test_batch():
    x1 = torch.tensor([1, 2, 3], dtype=torch.float)
    e1 = torch.tensor([[0, 1, 1, 2], [1, 0, 2, 1]])
    x2 = torch.tensor([1, 2], dtype=torch.float)
    e2 = torch.tensor([[0, 1], [1, 0]])

    data = Batch.from_data_list([Data(x1, e1), Data(x2, e2)])

    assert data.__repr__() == 'Batch(batch=[5], edge_index=[2, 6], x=[5])'
    assert len(data) == 3
    assert data.x.tolist() == [1, 2, 3, 1, 2]
    assert data.edge_index.tolist() == [[0, 1, 1, 2, 3, 4], [1, 0, 2, 1, 4, 3]]
    assert data.batch.tolist() == [0, 0, 0, 1, 1]
    assert data.num_graphs == 2



from pytest import approx
import torch
from torch_geometric.transforms import RandomRotate
from torch_geometric.data import Data


def test_spherical():
    assert RandomRotate(-180).__repr__() == 'RandomRotate((-180, 180))'

    pos = torch.tensor([[1, 0], [0, 1]], dtype=torch.float)
    data = Data(pos=pos)

    out = RandomRotate((90, 90))(data).pos.view(-1).tolist()
    assert approx(out) == [0, 1, -1, 0]



import torch

from .num_nodes import maybe_num_nodes


def degree(index, num_nodes=None, dtype=None):
    """Computes the degree of a given index tensor.

    Args:
        index (LongTensor): Source or target indices of edges.
        num_nodes (int, optional): The number of nodes in :attr:`index`.
            (default: :obj:`None`)
        dtype (:obj:`torch.dtype`, optional): The desired data type of returned
            tensor.

    :rtype: :class:`Tensor`

    .. testsetup::

        import torch

    .. testcode::

        from torch_geometric.utils import degree
        index = torch.tensor([0, 1, 0, 2, 0])
        output = degree(index)
        print(output)

    .. testoutput::

       tensor([3., 1., 1.])
    """

    num_nodes = maybe_num_nodes(index, num_nodes)
    out = torch.zeros((num_nodes), dtype=dtype, device=index.device)
    return out.scatter_add_(0, index, out.new_ones((index.size(0))))



from torch_geometric.utils import scatter_


def global_add_pool(x, batch, size=None):
    r"""Returns batch-wise graph-level-outputs by adding node features
    across the node node dimension, so that for a single graph
    :math:`\mathcal{G}_i` its output is computed by

    .. math::
        \mathbf{r}_i = \sum_{n=1}^{N_i} \mathbf{x}_n

    Args:
        x (Tensor): Node feature matrix
            :math:`\mathbf{X} \in \mathbb{R}^{(N_1 + \ldots + N_B) \times F}`.
        batch (LongTensor): Batch vector :math:`\mathbf{b} \in {\{ 0, \ldots,
            B-1\}}^N`, which assigns each node to a specific example.
        size (int, optional): Batch-size :math:`B`.
            Automatically calculated if not given. (default: :obj:`None`)

    :rtype: :class:`Tensor`
    """

    size = batch[-1].item() + 1 if size is None else size
    return scatter_('add', x, batch, dim_size=size)


def global_mean_pool(x, batch, size=None):
    r"""Returns batch-wise graph-level-outputs by averaging node features
    across the node node dimension, so that for a single graph
    :math:`\mathcal{G}_i` its output is computed by

    .. math::
        \mathbf{r}_i = \frac{1}{N_i} \sum_{n=1}^{N_i} \mathbf{x}_n

    Args:
        x (Tensor): Node feature matrix
            :math:`\mathbf{X} \in \mathbb{R}^{(N_1 + \ldots + N_B) \times F}`.
        batch (LongTensor): Batch vector :math:`\mathbf{b} \in {\{ 0, \ldots,
            B-1\}}^N`, which assigns each node to a specific example.
        size (int, optional): Batch-size :math:`B`.
            Automatically calculated if not given. (default: :obj:`None`)

    :rtype: :class:`Tensor`
    """

    size = batch[-1].item() + 1 if size is None else size
    return scatter_('mean', x, batch, dim_size=size)


def global_max_pool(x, batch, size=None):
    r"""Returns batch-wise graph-level-outputs by taking the channel-wise
    maximum node features across the node dimension, so that for a single graph
    :math:`\mathcal{G}_i` its output is computed by

    .. math::
        \mathbf{r}_i = \mathrm{max}_{n=1}^{N_i} \, \mathbf{x}_n

    Args:
        x (Tensor): Node feature matrix
            :math:`\mathbf{X} \in \mathbb{R}^{(N_1 + \ldots + N_B) \times F}`.
        batch (LongTensor): Batch vector :math:`\mathbf{b} \in {\{ 0, \ldots,
            B-1\}}^N`, which assigns each node to a specific example.
        size (int, optional): Batch-size :math:`B`.
            Automatically calculated if not given. (default: :obj:`None`)

    :rtype: :class:`Tensor`
    """

    size = batch[-1].item() + 1 if size is None else size
    return scatter_('max', x, batch, dim_size=size)



# import torch
# from torch_geometric.nn.pool import graclus
# from torch_geometric.data import Batch

# def test_graclus_pool():
#     x = torch.Tensor([[1, 6], [2, 5], [3, 4], [4, 3], [5, 2], [6, 1]])
#     row = torch.LongTensor([0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5])
#     col = torch.LongTensor([1, 2, 0, 2, 0, 1, 4, 5, 3, 5, 3, 4])
#     edge_index = torch.stack([row, col], dim=0)
#     batch = torch.LongTensor([0, 0, 0, 1, 1, 1])
#     data = Batch(x=x, edge_index=edge_index, batch=batch)

#     data = graclus_pool(data)

#     assert data.num_nodes == 4



import torch
from torch_geometric.data import Data
from torch_sparse import coalesce


def test_data():
    x = torch.tensor([[1, 3, 5], [2, 4, 6]], dtype=torch.float).t()
    edge_index = torch.tensor([[0, 0, 1, 1, 2], [1, 1, 0, 2, 1]])
    data = Data(x=x, edge_index=edge_index).to(torch.device('cpu'))

    N = data.num_nodes

    assert data.x.tolist() == x.tolist()
    assert data['x'].tolist() == x.tolist()

    assert sorted(data.keys) == ['edge_index', 'x']
    assert len(data) == 2
    assert 'x' in data and 'edge_index' in data and 'pos' not in data

    assert data.cat_dim('x', data.x) == 0
    assert data.cat_dim('edge_index', data.edge_index) == -1

    assert not data.x.is_contiguous()
    data.contiguous()
    assert data.x.is_contiguous()

    assert not data.is_coalesced()
    data.edge_index, _ = coalesce(data.edge_index, None, N, N)
    assert data.is_coalesced()

    data['x'] = x + 1
    assert data.x.tolist() == (x + 1).tolist()

    assert data.__repr__() == 'Data(edge_index=[2, 4], x=[3, 2])'

    dictionary = {'x': data.x, 'edge_index': data.edge_index}
    data = Data.from_dict(dictionary)
    assert sorted(data.keys) == ['edge_index', 'x']

    assert not data.contains_isolated_nodes()
    assert not data.contains_self_loops()
    assert data.is_undirected()
    assert not data.is_directed()

    assert data.num_nodes == 3
    assert data.num_edges == 4
    assert data.num_features == 2

    data.x = None
    assert data.num_nodes == 3

    data.edge_index = None
    assert data.num_nodes is None
    assert data.num_edges is None



class Center(object):
    r"""Centers node positions around the origin.

    .. testsetup::

        import torch
        from torch_geometric.data import Data

    .. testcode::

        from torch_geometric.transforms import Center

        pos = torch.Tensor([[0, 0], [2, 0], [4, 0]])
        data = Data(pos=pos)

        data = Center()(data)
    """

    def __call__(self, data):
        data.pos = data.pos - data.pos.mean(dim=0, keepdim=True)
        return data

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)



import torch
from torch_geometric.utils import one_hot


def test_one_hot():
    src = torch.LongTensor([1, 0, 3])
    expected_output = [[0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 0, 1]]

    assert one_hot(src).tolist() == expected_output
    assert one_hot(src, 4).tolist() == expected_output

    src = torch.LongTensor([[1, 0], [0, 1], [2, 0]])
    expected_output = [[0, 1, 0, 1, 0], [1, 0, 0, 0, 1], [0, 0, 1, 1, 0]]
    assert one_hot(src).tolist() == expected_output
    assert one_hot(src, torch.tensor([3, 2])).tolist() == expected_output



from math import exp

import torch
from torch.autograd import Variable
from torch_geometric.utils import softmax


def test_softmax():
    row = torch.LongTensor([0, 0, 0, 1, 1, 1])
    col = torch.LongTensor([1, 2, 3, 0, 2, 3])
    edge_attr = torch.Tensor([0, 1, 2, 0, 1, 2])
    e = [exp(1), exp(2), exp(3)]
    e_sum = e[0] + e[1] + e[2]
    e = [e[0] / e_sum, e[1] / e_sum, e[2] / e_sum]

    output = softmax(edge_attr, row)

    output = softmax(Variable(edge_attr), row)

    output = softmax(edge_attr, col)

    edge_attr = torch.Tensor([[1, 1], [2, 2], [3, 3], [1, 1], [2, 2], [3, 3]])
    output = softmax(edge_attr, row)

    output = softmax(Variable(edge_attr), row)
    output



import torch
from torch_geometric.transforms import FaceToEdge
from torch_geometric.data import Data


def test_face_to_edge():
    assert FaceToEdge().__repr__() == 'FaceToEdge()'

    face = torch.tensor([[0, 0], [1, 1], [2, 3]])
    data = Data()
    data.face = face

    row, col = FaceToEdge()(data).edge_index
    assert row.tolist() == [0, 0, 0, 1, 1, 1, 2, 2, 3, 3]
    assert col.tolist() == [1, 2, 3, 0, 2, 3, 0, 1, 0, 1]



import torch
from torch_geometric.utils import (contains_self_loops, remove_self_loops,
                                   add_self_loops)


def test_contains_self_loops():
    row = torch.tensor([0, 1, 0])
    col = torch.tensor([1, 0, 0])

    assert contains_self_loops(torch.stack([row, col], dim=0))

    row = torch.tensor([0, 1, 1])
    col = torch.tensor([1, 0, 2])

    assert not contains_self_loops(torch.stack([row, col], dim=0))


def test_remove_self_loops():
    row = torch.tensor([1, 0, 1, 0, 2, 1])
    col = torch.tensor([0, 1, 1, 1, 2, 0])
    edge_index = torch.stack([row, col], dim=0)
    edge_attr = [[1, 2], [3, 4], [5, 6], [7, 8], [9, 10], [11, 12]]
    edge_attr = torch.tensor(edge_attr)

    out = remove_self_loops(edge_index, edge_attr)
    assert out[0].tolist() == [[1, 0, 0, 1], [0, 1, 1, 0]]
    assert out[1].tolist() == [[1, 2], [3, 4], [7, 8], [11, 12]]


def test_add_self_loops():
    row = torch.tensor([0, 1, 0])
    col = torch.tensor([1, 0, 0])
    edge_index = torch.stack([row, col], dim=0)

    expected = [[0, 1, 0, 0, 1], [1, 0, 0, 0, 1]]
    assert add_self_loops(edge_index).tolist() == expected



import random


class RandomFlip(object):
    """Flips node positions along a given axis randomly with a given
    probability.

    Args:
        axis (int): The axis along the position of nodes being flipped.
        p (float, optional): Probability that node positions will be flipped.
            (default: :obj:`0.5`)

    .. testsetup::

        import torch
        from torch_geometric.data import Data

    .. testcode::

        from torch_geometric.transforms import RandomFlip

        pos = torch.tensor([[-1, 1], [-3, 0], [2, -1]], dtype=torch.float)
        data = Data(pos=pos)

        data = RandomFlip(axis=0, p=1)(data)

        print(data.pos)

    .. testoutput::

        tensor([[ 1.,  1.],
                [ 3.,  0.],
                [-2., -1.]])
    """

    def __init__(self, axis, p=0.5):
        self.axis = axis
        self.p = p

    def __call__(self, data):
        if random.random() < self.p:
            data.pos[:, self.axis] = -data.pos[:, self.axis]
        return data

    def __repr__(self):
        return '{}(axis={}, p={})'.format(self.__class__.__name__, self.axis,
                                          self.p)



import torch
from torch_geometric.transforms import Spherical
from torch_geometric.data import Data


def test_spherical():
    assert Spherical().__repr__() == 'Spherical(norm=True, max_value=None)'

    pos = torch.tensor([[0, 0, 0], [0, 1, 1]], dtype=torch.float)
    edge_index = torch.tensor([[0, 1], [1, 0]])
    data = Data(edge_index=edge_index, pos=pos)
    print(data)

    # out = Spherical()(data).edge_attr.view(-1).tolist()
    # assert approx(out) == [1, 0.25, 0, 1, 0.75, 1]

    # data.edge_attr = torch.tensor([1, 1], dtype=torch.float)
    # out = Spherical()(data).edge_attr.view(-1).tolist()
    # assert approx(out) == [1, 1, 0.25, 0, 1, 1, 0.75, 1]



import torch
from torch.nn import Parameter
from torch_geometric.utils import remove_self_loops, scatter_

from ..inits import uniform


class GraphConv(torch.nn.Module):
    r"""The graph neural network operator from the `"Weisfeiler and Leman Go
    Neural: Higher-order Graph Neural Networks"
    <https://arxiv.org/abs/1810.02244>`_ paper

    .. math::
        \mathbf{x}^{\prime}_i = \mathbf{\Theta} \mathbf{x}_i +
        \sum_{j \in \mathcal{N}(i)} \mathbf{\Theta} \mathbf{x}_j.

    Args:
        in_channels (int): Size of each input sample.
        out_channels (int): Size of each output sample.
        aggr (string): The aggregation operator to use (one of :obj:`"add"`,
            :obj:`"mean"`, :obj:`"max"`). (default: :obj:`"add"`)
        bias (bool, optional): If set to :obj:`False`, the layer will not learn
            an additive bias. (default: :obj:`True`)
    """

    def __init__(self, in_channels, out_channels, aggr='add', bias=True):
        super(GraphConv, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.aggr = aggr
        self.weight = Parameter(torch.Tensor(in_channels, out_channels))
        self.root = Parameter(torch.Tensor(in_channels, out_channels))

        if bias:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        size = self.in_channels
        uniform(size, self.weight)
        uniform(size, self.root)
        uniform(size, self.bias)

    def forward(self, x, edge_index):
        """"""
        x = x.unsqueeze(-1) if x.dim() == 1 else x
        edge_index, _ = remove_self_loops(edge_index)
        row, col = edge_index

        out = torch.mm(x, self.weight)
        out = scatter_(self.aggr, out[col], row, dim_size=x.size(0))
        out = out + torch.mm(x, self.root)

        if self.bias is not None:
            out = out + self.bias

        return out

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__, self.in_channels,
                                   self.out_channels)



import torch
from torch_geometric.utils import degree


class TargetIndegree(object):
    r"""Saves the globally normalized degree of target nodes (mapped to the
    fixed interval :math:`[0, 1]`)

    .. math::

        \mathbf{u}(i,j) = \frac{\deg(j)}{\max_{v \in \mathcal{V}} \deg(v)}

    in its edge attributes.

    Args:
        cat (bool, optional): Concat pseudo-coordinates to edge attributes
            instead of replacing them. (default: :obj:`True`)

    .. testsetup::

        import torch
        from torch_geometric.data import Data

    .. testcode::

        from torch_geometric.transforms import TargetIndegree

        edge_index = torch.tensor([[0, 1, 1, 2], [1, 0, 2, 1]])
        data = Data(edge_index=edge_index)

        data = TargetIndegree()(data)

        print(data.edge_attr)

    .. testoutput::

        tensor([[1.0000],
                [0.5000],
                [0.5000],
                [1.0000]])
    """

    def __init__(self, cat=True):
        self.cat = cat

    def __call__(self, data):
        col, pseudo = data.edge_index[1], data.edge_attr

        deg = degree(col, data.num_nodes)
        deg = deg / deg.max()
        deg = deg[col]
        deg = deg.view(-1, 1)

        if pseudo is not None and self.cat:
            pseudo = pseudo.view(-1, 1) if pseudo.dim() == 1 else pseudo
            data.edge_attr = torch.cat([pseudo, deg.type_as(pseudo)], dim=-1)
        else:
            data.edge_attr = deg

        return data

    def __repr__(self):
        return '{}(cat={})'.format(self.__class__.__name__, self.cat)



import torch
from torch_geometric.transforms import RadiusGraph
from torch_geometric.data import Data


def test_radius_graph():
    assert RadiusGraph(1).__repr__() == 'RadiusGraph(r=1)'

    pos = [[0, 0], [1, 0], [2, 0], [0, 1], [-2, 0], [0, -2]]
    pos = torch.tensor(pos, dtype=torch.float)
    data = Data(pos=pos)

    edge_index = RadiusGraph(1)(data).edge_index
    assert edge_index.tolist() == [[0, 0, 1, 1, 2, 3], [1, 3, 0, 2, 1, 0]]



import torch.utils.data
from torch.utils.data.dataloader import default_collate

from torch_geometric.data import Batch


class DataLoader(torch.utils.data.DataLoader):
    def __init__(self, dataset, batch_size=1, shuffle=True, **kwargs):
        super(DataLoader, self).__init__(
            dataset,
            batch_size,
            shuffle,
            collate_fn=lambda batch: Batch.from_data_list(batch),
            **kwargs)


class DenseDataLoader(torch.utils.data.DataLoader):
    def __init__(self, dataset, batch_size=1, shuffle=True, **kwargs):
        def dense_collate(data_list):
            batch = Batch()
            for key in data_list[0].keys:
                batch[key] = default_collate([d[key] for d in data_list])
            return batch

        super(DenseDataLoader, self).__init__(
            dataset, batch_size, shuffle, collate_fn=dense_collate, **kwargs)



import sys
import random
import os.path as osp
import shutil

from torch_geometric.datasets import Planetoid
from torch_geometric.data import DataLoader


def test_citeseer():
    root = osp.join('/', 'tmp', str(random.randrange(sys.maxsize)), 'test')
    dataset = Planetoid(root, 'Citeseer')
    loader = DataLoader(dataset, batch_size=len(dataset))

    assert len(dataset) == 1
    assert dataset.__repr__() == 'Citeseer()'

    for data in loader:
        assert data.num_graphs == 1
        assert data.num_nodes == 3327
        assert data.num_edges / 2 == 4552

        assert len(data) == 7
        assert list(data.x.size()) == [data.num_nodes, 3703]
        assert list(data.y.size()) == [data.num_nodes]
        assert data.y.max() + 1 == 6
        assert data.train_mask.sum() == 6 * 20
        assert data.val_mask.sum() == 500
        assert data.test_mask.sum() == 1000
        assert (data.train_mask & data.val_mask & data.test_mask).sum() == 0
        assert list(data.batch.size()) == [data.num_nodes]

        assert data.contains_isolated_nodes()
        assert not data.contains_self_loops()
        assert data.is_undirected()

    shutil.rmtree(root)



import torch
import networkx as nx
from torch_geometric.data import InMemoryDataset, Data


class KarateClub(InMemoryDataset):
    def __init__(self, transform=None):
        super(KarateClub, self).__init__('.', transform, None, None)

        G = nx.karate_club_graph()
        adj = nx.to_scipy_sparse_matrix(G).tocoo()
        row = torch.from_numpy(adj.row).to(torch.long)
        col = torch.from_numpy(adj.col).to(torch.long)
        edge_index = torch.stack([row, col], dim=0)
        data = Data(edge_index=edge_index)
        data.x = torch.eye(data.num_nodes, dtype=torch.float)
        self.data, self.slices = self.collate([data])

    def _download(self):
        return

    def _process(self):
        return

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)



import tarfile
import zipfile


def maybe_log(path, log=True):
    if log:
        print('Extracting', path)


def extract_tar(path, folder, mode='r:gz', log=True):
    maybe_log(path, log)
    with tarfile.open(path, mode) as f:
        f.extractall(folder)


def extract_zip(path, folder, log=True):
    maybe_log(path, log)
    with zipfile.ZipFile(path, 'r') as f:
        f.extractall(folder)



import os.path as osp
from six.moves import urllib

from .makedirs import makedirs


def download_url(url, folder, log=True):
    if log:
        print('Downloading', url)

    makedirs(folder)

    data = urllib.request.urlopen(url)
    filename = url.rpartition('/')[2]
    path = osp.join(folder, filename)

    with open(path, 'wb') as f:
        f.write(data.read())

    return path



import torch
from torch_geometric.utils import normalized_cut


def test_normalized_cut():
    row = torch.LongTensor([0, 1, 1, 1, 2, 2, 3, 3, 4, 4])
    col = torch.LongTensor([1, 0, 2, 3, 1, 4, 1, 4, 2, 3])
    edge_attr = torch.Tensor([3, 3, 6, 3, 6, 1, 3, 2, 1, 2])
    expected_output = [4, 4, 5, 2.5, 5, 1, 2.5, 2, 1, 2]

    output = normalized_cut(torch.stack([row, col], dim=0), edge_attr)
    assert output.tolist() == expected_output
